/*******************************************************************************
 * Copyright (c) 2000, 2004 IBM Corporation and others.
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 * 
 * Contributors:
 *     IBM Corporation - initial API and implementation
 *******************************************************************************/
package org.eclipse.target.internal.ui;

import java.util.*;
import java.util.List;

import org.eclipse.core.runtime.CoreException;
import org.eclipse.jface.action.*;
import org.eclipse.jface.viewers.*;
import org.eclipse.jface.wizard.WizardDialog;
import org.eclipse.swt.SWT;
import org.eclipse.swt.custom.SashForm;
import org.eclipse.swt.dnd.FileTransfer;
import org.eclipse.swt.dnd.Transfer;
import org.eclipse.swt.events.*;
import org.eclipse.swt.layout.GridData;
import org.eclipse.swt.layout.GridLayout;
import org.eclipse.swt.widgets.*;
import org.eclipse.target.*;
import org.eclipse.target.internal.*;
import org.eclipse.target.internal.ui.actions.CreateNewFolderAction;
import org.eclipse.target.internal.ui.dnd.*;
import org.eclipse.ui.*;
import org.eclipse.ui.actions.ActionFactory;
import org.eclipse.ui.dialogs.PropertyDialogAction;
import org.eclipse.ui.internal.WorkbenchImages;
import org.eclipse.ui.model.BaseWorkbenchContentProvider;
import org.eclipse.ui.model.WorkbenchLabelProvider;
import org.eclipse.ui.part.PluginTransfer;
import org.eclipse.ui.part.ViewPart;

/**
 * Is a view that allows browsing remote target sites. It is modeled after
 * a file explorer: a tree of folders is show with a table of the folder's
 * contents.
 * <p>
 * Progress is shown in the main workbench window's status line progress 
 * monitor.</p>
 * 
 * @see Site
 * @see IRemoteTargetResource
 */
public class SiteExplorerView extends ViewPart implements ITargetListener {

	// The tree viewer showing the folders and sites
	private TreeViewer folderTree;
	
	// The table view that shows the resources in the currently selected folder
	// from the folders tree.
	private TableViewer folderContentsTable;
	
	// The root
	private Object root;
	
	// The view's actions
	private Action addSiteAction;
	private Action newFolderAction;
	private PropertyDialogAction propertiesAction;
	
	/**
	 * Sorter for the folderContents table
	 */
	class FolderListingSorter extends ViewerSorter {
		private boolean reversed = false;
		private int columnNumber;
		
		public static final int NAME = 0;
		public static final int SIZE = 1;
		public static final int MODIFIED = 2;
				
		// column headings:	"Name" "Size" "Modified"
		private int[][] SORT_ORDERS_BY_COLUMN = {
			{NAME},	/* name */ 
			{SIZE, NAME},	/* size */
			{MODIFIED, NAME, SIZE},	/* modified */
		};
		
		public FolderListingSorter(int columnNumber) {
			this.columnNumber = columnNumber;
		}
		
		public int compare(Viewer viewer, Object o1, Object o2) {
			ITargetResource e1 = (ITargetResource)o1;
			ITargetResource e2 = (ITargetResource)o2;
			int[] columnSortOrder = SORT_ORDERS_BY_COLUMN[columnNumber];
			int result = 0;
			for (int i = 0; i < columnSortOrder.length; ++i) {
				result = compareColumnValue(columnSortOrder[i], e1, e2);
				if (result != 0)
					break;
			}
			if (reversed)
				result = -result;
			return result;
		}
		
		int compareColumnValue(int columnNumber, ITargetResource r1, ITargetResource r2) {
			switch (columnNumber) {
				case NAME:
					if (r1.isDirectory() && r2.isDirectory())
						return compareNames(r1, r2);
					else if (r1.isDirectory())
						return -1;
					else if (r2.isDirectory())
						return 1;
					return compareNames(r1, r2);
				case SIZE:
					return new Long(r1.getSize()).compareTo(new Long(r2.getSize()));
				case MODIFIED:
					try {
						return getCollator().compare(r1.getProperty(ITargetResource.LAST_MODIFIED_TIMESTAMP_PROPERTY), r2.getProperty(ITargetResource.LAST_MODIFIED_TIMESTAMP_PROPERTY));
					} catch (CoreException e) {
						TargetPlugin.log(e);
					}
				default:
					return 0;
			}
		}
		
		protected int compareNames(ITargetResource resource1, ITargetResource resource2) {
			return resource1.getName().compareTo(resource2.getName());
		}
		
		/**
		 * Returns the number of the column by which this is sorting.
		 */
		public int getColumnNumber() {
			return columnNumber;
		}
		
		/**
		 * Returns true for descending, or false
		 * for ascending sorting order.
		 */
		public boolean isReversed() {
			return reversed;
		}
		
		/**
		 * Sets the sorting order.
		 */
		public void setReversed(boolean newReversed) {
			reversed = newReversed;
		}
	}
		
	/**
	 * @see IWorkbenchPart#createPartControl(Composite)
	 */
	public void createPartControl(Composite top) {
		Composite p = new Composite(top, SWT.NULL);
		GridData data = new GridData (GridData.FILL_BOTH);		
		p.setLayoutData(data);
		GridLayout gridLayout = new GridLayout();
		gridLayout.numColumns = 3;
		gridLayout.marginHeight = 0;
		gridLayout.marginWidth = 0;
		p.setLayout (gridLayout);
		
		SashForm sash = new SashForm(p, SWT.HORIZONTAL);
		data = new GridData(GridData.FILL_BOTH);
		data.horizontalSpan = 3;
		sash.setLayoutData(data);
		
		folderTree = createFolderTree(sash);

		// Create the folders contents viewer
		folderContentsTable = createFolderContentsTable(sash);

		sash.setWeights(new int[] {33, 67});
		
		Target.getSiteManager().addSiteListener(this);
		
		root = Target.getSiteManager();
		initalizeActions();
		folderTree.setInput(root);
		
		// TODO: F1 Help
		// WorkbenchHelp.setHelp(folderTree.getControl(), IHelpContextIds.SITE_EXPLORER_VIEW);
	}

	/**
	 * @param sash
	 */
	private TreeViewer createFolderTree(SashForm sash) {
		final TreeViewer folderTree = createTargetFolderViewer(sash, true);
		folderTree.getControl().addKeyListener(new KeyAdapter() {
			public void keyPressed(KeyEvent event) {
				if (event.keyCode == SWT.F5) {
					IStructuredSelection s = (IStructuredSelection)folderTree.getSelection();
					if(s.isEmpty()) {
						folderTree.refresh();
					} else {
						Object[] elements = s.toArray();
						for (int i = 0; i < elements.length; i++) {
							Object object = elements[i];
							folderTree.refresh(object);
						}
						ITargetResource[] resources = getSelectedRemoteFolder(s);
						if (resources.length > 0) {
							updateFileTable(resources[0]);
						}
					}
				}
			}
		});
		folderTree.addSelectionChangedListener(new ISelectionChangedListener() {
			public void selectionChanged(SelectionChangedEvent event) {
				IStructuredSelection selection = (IStructuredSelection)folderTree.getSelection();
				final ITargetResource[] remoteFolders = getSelectedRemoteFolder(selection);
				if(remoteFolders.length == 1) {
					updateFileTable(remoteFolders[0]);
				}
			}
		});
		folderTree.addDoubleClickListener(new IDoubleClickListener() {
			public void doubleClick(DoubleClickEvent e) {
				IStructuredSelection selection = (IStructuredSelection)folderTree.getSelection();
				if(selection.size() == 1) {
					expandInTreeCurrentSelection(selection, true /*toggle expanded*/);
				}
			}
		});
		
		// Add drop support to tree
		int ops = TargetResourceDragListener.getSupportedOperations();
		Transfer[] dragTransfers = new Transfer[] { TargetResourceTransfer.getInstance(), PluginTransfer.getInstance()};
		Transfer[] dropTransfers = new Transfer[] { TargetResourceTransfer.getInstance(), PluginTransfer.getInstance(), FileTransfer.getInstance()};
		folderTree.addDragSupport(ops, dragTransfers, new TargetResourceDragListener(folderTree));
		folderTree.addDropSupport(ops, dropTransfers, new TargetResourceTreeDropAdapter(this, folderTree));
		
		return folderTree;
	}

	private TableViewer createFolderContentsTable(SashForm sash) {
		Table table = new Table(sash, SWT.FULL_SELECTION | SWT.H_SCROLL | SWT.V_SCROLL);
		TableLayout layout = new TableLayout();
		table.setLayout(layout);
		table.setHeaderVisible(true);
		
		TableColumn tableColumn = new TableColumn(table, SWT.NULL);
		tableColumn.setText(TargetMessages.SiteExplorerView_Name_1); 
		tableColumn.addSelectionListener(getColumnListener());
		layout.addColumnData(new ColumnWeightData(30, true));

		tableColumn = new TableColumn(table, SWT.NULL);
		tableColumn.setText(TargetMessages.SiteExplorerView_Size_2); 
		tableColumn.setAlignment(SWT.RIGHT);
		tableColumn.addSelectionListener(getColumnListener());
		layout.addColumnData(new ColumnWeightData(10, true));

		tableColumn = new TableColumn(table, SWT.NULL);
		tableColumn.setText(TargetMessages.SiteExplorerView_Modified_3); 
		tableColumn.addSelectionListener(getColumnListener());
		layout.addColumnData(new ColumnWeightData(30, true));

		tableColumn = new TableColumn(table, SWT.NULL);
		tableColumn.setText(TargetMessages.SiteExplorerView_URL_4); 
		layout.addColumnData(new ColumnWeightData(30, true));

		final TableViewer folderContentsTable = new TableViewer(table);
		folderContentsTable.setContentProvider(new BaseWorkbenchContentProvider());
		folderContentsTable.setLabelProvider(new SiteExplorerViewLabelProvider());
		
		folderContentsTable.getControl().addKeyListener(new KeyAdapter() {
			public void keyPressed(KeyEvent event) {
				if (event.keyCode == SWT.F5) {
					ITargetResource folder = (ITargetResource)folderContentsTable.getInput();
					if(folder != null) {
						folderContentsTable.refresh();
					}
				}
			}
		});
		
		folderContentsTable.addDoubleClickListener(new IDoubleClickListener() {
			public void doubleClick(DoubleClickEvent e) {
				IStructuredSelection selection = (IStructuredSelection)folderContentsTable.getSelection();
				if(selection.size() == 1) {
					final ITargetResource[] remoteFolders = getSelectedRemoteFolder(selection);
					if(remoteFolders.length == 1) {
						IStructuredSelection treeSelection = (IStructuredSelection)folderTree.getSelection();
						expandInTreeCurrentSelection(treeSelection, false /*don't toggle*/);
						folderTree.setSelection(new StructuredSelection(remoteFolders[0]));
					}
				}
			}
		});
		FolderListingSorter sorter = new FolderListingSorter(FolderListingSorter.NAME);
		sorter.setReversed(false);
		folderContentsTable.setSorter(sorter);
		
		// Add drap and drop support
		int ops = TargetResourceDragListener.getSupportedOperations();
		Transfer[] transfers = new Transfer[] { TargetResourceTransfer.getInstance(), PluginTransfer.getInstance()};
		folderContentsTable.addDragSupport(ops, transfers, new TargetResourceDragListener(folderContentsTable));
		Transfer[] dropTransfers = new Transfer[] { TargetResourceTransfer.getInstance(), PluginTransfer.getInstance(), FileTransfer.getInstance()};
		folderContentsTable.addDropSupport(ops, dropTransfers, new TargetResourceViewerDropAdapter(this, folderContentsTable));

		return folderContentsTable;
	}

	public static TreeViewer createTargetFolderViewer(Composite composite, boolean multiSelect) {
		
		TreeViewer folderTree = new TreeViewer(composite, (multiSelect ? SWT.MULTI : SWT.SINGLE) | SWT.H_SCROLL | SWT.V_SCROLL);
		folderTree.setContentProvider(new DeferredContentProvider());
		folderTree.setLabelProvider(new WorkbenchLabelProvider());
		folderTree.setSorter(new SiteViewSorter());
		folderTree.setSorter(new ViewerSorter() {
			public int compare(Viewer viewer, Object e1, Object e2) {
				String name1 = ""; //$NON-NLS-1$
				String name2 = ""; //$NON-NLS-1$
				if(e1 instanceof ITargetResource) {
					name1 = ((ITargetResource)e1).getName();
				} else if(e1 instanceof Site) {
					name1 = ((Site)e1).getLocation();
				}
				if(e2 instanceof ITargetResource) {
					name2 = ((ITargetResource)e2).getName();
				} else if(e2 instanceof Site) {
					name2 = ((Site)e2).getLocation();
				}
				
				return getCollator().compare(name1, name2);
			}
		});
		
		// show only folders in tree
		folderTree.addFilter(new ViewerFilter() {
			public boolean select(Viewer viewer, Object parentElement, Object element) {
				if(element instanceof ITargetResource) {
					return ((ITargetResource)element).isDirectory();
				}
				return true;
			}
		});
		return folderTree;
	}

	private Shell getShell() {
		return folderTree.getTree().getShell();
	}

	private ITargetResource[] getSelectedRemoteFolder(IStructuredSelection selection) {		
		if (!selection.isEmpty()) {
			final List folders = new ArrayList();
			Iterator it = selection.iterator();
			while(it.hasNext()) {
				Object o = it.next();
				if(o instanceof ITargetResource) {
					folders.add(o);
				} else if (o instanceof Site) {
					folders.add(((Site)o).getRootResource());
				}
			}
			return (ITargetResource[]) folders.toArray(new ITargetResource[folders.size()]);
		}
		return new ITargetResource[0];
	}
	
	private void expandInTreeCurrentSelection(IStructuredSelection selection, boolean toggle) {
		if (!selection.isEmpty()) {
			Iterator it = selection.iterator();
			while(it.hasNext()) {
				Object element = it.next();
				if(toggle) {
					folderTree.setExpandedState(element, !folderTree.getExpandedState(element));
				} else {
					folderTree.setExpandedState(element, true);
				}
			}
		}
	}

	/**
	 * Method updateFileTable.
	 */
	private void updateFileTable(ITargetResource remoteFolder) {
		if(remoteFolder != null && !remoteFolder.equals(folderContentsTable.getInput())) {
			folderContentsTable.setInput(remoteFolder);
		}
	}

	private void initalizeActions() {
		final Shell shell = folderContentsTable.getTable().getShell();
		// Create actions
		
		// Refresh (toolbar)
		addSiteAction = new Action(TargetMessages.SiteExplorerView_addSiteAction, TargetPlugin.getDefault().getImageDescriptor(ITargetImages.IMG_SITE_ELEMENT)) { 
			public void run() {
				CreateSiteWizard wizard = new CreateSiteWizard();
				WizardDialog dialog = new WizardDialog(shell, wizard);
				dialog.open();
			}
		};
		addSiteAction.setToolTipText(TargetMessages.SiteExplorerView_addSiteActionTooltip); 
		//TODO: F1 help
		// WorkbenchHelp.setHelp(addSiteAction, IHelpContextIds.ADD_SITE_ACTION);
		
		newFolderAction = new Action(TargetMessages.SiteExplorerView_newFolderAction, WorkbenchImages.getImageDescriptor(org.eclipse.ui.ISharedImages.IMG_OBJ_FOLDER)) { 
			public void run() {
				final Shell shell = folderTree.getTree().getShell();
				try {
					// assume that only one folder is selected in the folder tree, this
					// is enforced by isEnable() method for this action
					IStructuredSelection selection = (IStructuredSelection)folderTree.getSelection();
					Object currentSelection = selection.getFirstElement();
					
					ITargetResource selectedFolder;
					if(!selection.isEmpty()) {
						selectedFolder = getSelectedRemoteFolder(selection)[0];
					} else {
						selectedFolder = (ITargetResource)folderContentsTable.getInput();
					}
										
					ITargetResource newFolder = CreateNewFolderAction.createDir(shell, selectedFolder, TargetMessages.CreateNewFolderAction_newFolderName); 
					if (newFolder == null)
						return;
					
					// select the newly added folder
					folderTree.refresh(currentSelection);
					expandInTreeCurrentSelection(new StructuredSelection(currentSelection), false);
					folderTree.setSelection(new StructuredSelection(newFolder));
				} catch (CoreException e) {
					Utils.handle(getShell(), e);
					return;
				}
			}
			public boolean isEnabled() {
				return folderContentsTable.getInput() != null ||
		 				getSelectedRemoteFolder((IStructuredSelection)folderTree.getSelection()).length == 1;
			}				
		};
		// TODO: WorkbenchHelp.setHelp(newFolderAction, IHelpContextIds.NEW_FOLDER_ACTION);

		IActionBars bars = getViewSite().getActionBars();
		IToolBarManager tbm = bars.getToolBarManager();
		tbm.add(addSiteAction);
		tbm.update(false);
		
		// Properties
		propertiesAction = new PropertyDialogAction(shell, folderTree);
		getViewSite().getActionBars().setGlobalActionHandler(ActionFactory.PROPERTIES.getId(), propertiesAction);		
		IStructuredSelection selection = (IStructuredSelection)folderTree.getSelection();
		if (selection.size() == 1 && selection.getFirstElement() instanceof Site) {
			propertiesAction.setEnabled(true);
		} else {
			propertiesAction.setEnabled(false);
		}
		folderTree.addSelectionChangedListener(new ISelectionChangedListener() {
			public void selectionChanged(SelectionChangedEvent event) {
				IStructuredSelection ss = (IStructuredSelection)event.getSelection();
				boolean enabled = ss.size() == 1 && ss.getFirstElement() instanceof Site;
				propertiesAction.setEnabled(enabled);
			}
		});
		
		MenuManager treeMgr = new MenuManager();
		MenuManager tableMgr = new MenuManager();
		Tree tree = folderTree.getTree();
		Table table = folderContentsTable.getTable();
		Menu treeMenu = treeMgr.createContextMenu(tree);
		Menu tableMenu = tableMgr.createContextMenu(table);
		IMenuListener menuListener = new IMenuListener() {
			public void menuAboutToShow(IMenuManager manager) {
				// Misc additions
				MenuManager sub = new MenuManager(TargetMessages.SiteExplorerView_newMenu, IWorkbenchActionConstants.GROUP_ADD); 
				sub.add(new Separator(IWorkbenchActionConstants.MB_ADDITIONS));
				manager.add(sub);
				manager.add(new Separator(IWorkbenchActionConstants.MB_ADDITIONS));
				sub.add(addSiteAction);
				sub.add(newFolderAction);
				manager.add(propertiesAction);
			}
		};
		treeMgr.addMenuListener(menuListener);
		treeMgr.setRemoveAllWhenShown(true);
		tableMgr.addMenuListener(menuListener);
		tableMgr.setRemoveAllWhenShown(true);
		tree.setMenu(treeMenu);
		table.setMenu(tableMenu);
		getSite().registerContextMenu(tableMgr, folderContentsTable);
		getSite().registerContextMenu(treeMgr, folderTree);
	}
	
	/**
	 * Add the new site to the viewer and make it the current selection.
	 * 
	 * @see ITargetListener#siteAdded(Site)
	 */
	public void siteAdded(Site site) {
		folderTree.add(root, site);
		folderTree.setSelection(new StructuredSelection(site));
	}

	/* (non-Javadoc)
	 * @see org.eclipse.target.ITargetListener#aboutToRemove(org.eclipse.target.Site, org.eclipse.swt.widgets.Shell)
	 */
	public boolean aboutToRemove(Site[] sites, Shell shell) {
		return true;
	}
	
	/**
	 * Remote the site from the viewer and select the next site in the
	 * tree.
	 * 
	 * @see ITargetListener#siteRemoved(Site)
	 */
	public void siteRemoved(Site site) {
		folderTree.remove(site);	
		selectNextObjectInTreeViewer();	
	}
	
	private void selectNextObjectInTreeViewer() {
		Object[] items = folderTree.getVisibleExpandedElements();
		if(items.length > 0) {
			folderTree.setSelection(new StructuredSelection(items[0]));
		} else {
			folderContentsTable.setInput(null);
		}
	}

	/**
	 * Adds the listener that sets the sorter.
	 */
	private SelectionListener getColumnListener() {
		/**
	 	 * This class handles selections of the column headers.
		 * Selection of the column header will cause resorting
		 * of the shown tasks using that column's sorter.
		 * Repeated selection of the header will toggle
		 * sorting order (ascending versus descending).
		 */
		return new SelectionAdapter() {
			/**
			 * Handles the case of user selecting the
			 * header area.
			 * <p>If the column has not been selected previously,
			 * it will set the sorter of that column to be
			 * the current tasklist sorter. Repeated
			 * presses on the same column header will
			 * toggle sorting order (ascending/descending).
			 */
			public void widgetSelected(SelectionEvent e) {
				// column selected - need to sort
				// only allow sorting on name for now
				int column = folderContentsTable.getTable().indexOf((TableColumn) e.widget);
				if(column == FolderListingSorter.NAME) {
					FolderListingSorter oldSorter = (FolderListingSorter)folderContentsTable.getSorter();
					if (oldSorter != null && column == oldSorter.getColumnNumber()) {
						oldSorter.setReversed(!oldSorter.isReversed());
						folderContentsTable.refresh();
					} else {
						folderContentsTable.setSorter(new FolderListingSorter(column));
					}
				}
			}
		};
	}

	/**
	 * @see IWorkbenchPart#setFocus()
	 */
	public void setFocus() {
	}
	/**
	 * @see org.eclipse.ui.IWorkbenchPart#dispose()
	 */
	public void dispose() {
		super.dispose();
		Target.getSiteManager().removeSiteListener(this);
	}

	/* private */ void handle(Shell shell, CoreException e) {
		Utils.handle(shell, e);
	}
	
}
