/*******************************************************************************
 * Copyright (c) 2000, 2004 IBM Corporation and others.
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 * 
 * Contributors:
 *     IBM Corporation - initial API and implementation
 *******************************************************************************/
package org.eclipse.team.internal.target.ui;

import java.net.URL;
import java.util.Iterator;

import org.eclipse.core.resources.IContainer;
import org.eclipse.core.runtime.*;
import org.eclipse.jface.dialogs.Dialog;
import org.eclipse.jface.resource.ImageDescriptor;
import org.eclipse.jface.viewers.*;
import org.eclipse.swt.SWT;
import org.eclipse.swt.layout.GridData;
import org.eclipse.swt.widgets.*;
import org.eclipse.target.ITargetResource;
import org.eclipse.target.Site;
import org.eclipse.target.internal.ui.DeferredContentProvider;
import org.eclipse.target.internal.ui.SiteViewSorter;
import org.eclipse.target.internal.ui.actions.CreateNewFolderAction;
import org.eclipse.team.internal.target.TeamTargetMessages;
import org.eclipse.team.internal.target.UrlUtil;
import org.eclipse.team.internal.ui.Utils;
import org.eclipse.ui.model.WorkbenchLabelProvider;

public class MappingSelectionPage extends TargetWizardPage {
	private IPath path = Path.EMPTY;
	private Site site;
	private IContainer container;
	private TreeViewer viewer;
	private boolean allowNewFolder = true;
	
	public MappingSelectionPage(String pageName, String title, String description, ImageDescriptor titleImage) {
		super(pageName, title, titleImage);
		setDescription(description);
	}

	public void setSite(Site site) {
		this.site = site;
	}
	
	public Site getSite() {
		return site;
	}
	
	public void createControl(Composite p) {
		Composite composite = createComposite(p, 1);
		// set F1 help
		//WorkbenchHelp.setHelp(composite, IHelpContextIds.TARGET_MAPPING_SELECTION_PAGE);
		
		createLabel(composite, TeamTargetMessages.MappingSelectionPage_label); 
		
		viewer = new TreeViewer(composite, SWT.BORDER | SWT.SINGLE);
		
		GridData data = new GridData (GridData.FILL_BOTH);
		viewer.getTree().setLayoutData(data);
		SiteViewSorter siteViewSorter = new SiteViewSorter();
		viewer.setContentProvider(new DeferredContentProvider());
		viewer.setLabelProvider(new WorkbenchLabelProvider());
        viewer.setSorter(siteViewSorter);
		viewer.addSelectionChangedListener(new ISelectionChangedListener() {
			public void selectionChanged(SelectionChangedEvent event) {
				updateTextPath();
			}
		});
		
		// include only folders in view
		viewer.addFilter(new ViewerFilter() {
			public boolean select(Viewer viewer, Object parentElement, Object element) {
				if(element instanceof ITargetResource) {
					return ((ITargetResource)element).isDirectory();
				}
				return true;
			}
		});
		
		if (allowNewFolder) {
			Button newFolderButton = new Button(composite, SWT.PUSH);
			newFolderButton.setText(TeamTargetMessages.MappingSelectionPage_newFolderLabel); 
			newFolderButton.addListener(SWT.Selection, new Listener() {
				public void handleEvent (Event event) {
					try {
						// assume that only one folder is selected in the folder tree
						IStructuredSelection selection = (IStructuredSelection)viewer.getSelection();
						Object currentSelection = selection.getFirstElement();
						final ITargetResource selectedFolder = getSelectedRemoteFolder(selection);					
						String defaultName = container.getName();
	
						ITargetResource newFolder = CreateNewFolderAction.createDir(getShell(), selectedFolder, defaultName);
						if (newFolder == null)
							return;
	
						viewer.refresh(currentSelection);
						viewer.setExpandedState(currentSelection, true);
						viewer.setSelection(new StructuredSelection(newFolder));
					} catch (CoreException e) {
						Utils.handle(e);
						return;
					}
				}			
			});
		}
		setViewerInput();
		setControl(composite);
		setPageComplete(true);
		Dialog.applyDialogFont(p);
	}
	
	private ITargetResource getSelectedRemoteFolder(IStructuredSelection selection) {		
		if (!selection.isEmpty()) {
			Iterator it = selection.iterator();
			while(it.hasNext()) {
				Object o = it.next();
				if(o instanceof ITargetResource) {
					return ((ITargetResource)o);
				} else if(o instanceof Site) {
					return ((Site)o).getRootResource();
				}
			}
		}
		return null;
	}
	
	/**
	 * Method updateTextPath.
	 */
	private void updateTextPath() {
		IStructuredSelection selection = (IStructuredSelection)viewer.getSelection();
		if (!selection.isEmpty()) {
			Iterator it = selection.iterator();
			while(it.hasNext()) {
				Object o = it.next();
				if(o instanceof ITargetResource) {
						ITargetResource element = (ITargetResource) o;
						URL remoteResourceURL;
						remoteResourceURL = element.toUrl();
						this.path = UrlUtil.getTrailingPath(
							remoteResourceURL,
							this.site.toUrl());
						setPageComplete(true);
					return;
				}
			}
		}
	}

	public IPath getMapping() {
		return this.path;
	}
	
	/**
	 * Attempt to set the viewer input.
	 * Do nothing if we don't have enough info yet to set it.
	 */
	private void setViewerInput() {
		if(this.site == null || viewer == null)
			return;
		viewer.setInput(new SiteViewerInput(new Site[] { site }));		
	}

	/**
	 * @see IDialogPage#setVisible(boolean)
	 */
	public void setVisible(boolean visible) {
		if(visible) {
			setViewerInput();
			viewer.setSelection(new StructuredSelection(site));
		}
		super.setVisible(visible);
	}
	
	/**
	 * @param container The container to set.
	 */
	public void setContainer(IContainer container) {
		this.container = container;
	}

	/**
	 * @param b
	 */
	public void setAllowNewFolder(boolean allowNewFolder) {
		this.allowNewFolder = allowNewFolder;
	}

}
